library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity EnigmaRAMTestbench is
    generic
	 (
		address_width_testbench 	: integer := 8;
		data_width_testbench 		: integer := 4
	 );
end EnigmaRAMTestbench;
 
architecture behavior OF EnigmaRAMTestbench is 
    component ram
	 generic
	 (
		address_width_ram 	: integer := 8;
		data_width_ram 		: integer := 4
	 );
    port(
         clock_ram 			: in  std_logic;
         data_in_ram 		: in  std_logic_vector(0 to data_width_ram-1);
         data_out_ram 		: out std_logic_vector(0 to data_width_ram-1);
         write_enable_ram 	: in  std_logic;
         write_address_ram : in  std_logic_vector(0 to address_width_ram-1);
         read_enable_ram 	: in  std_logic;
         read_address_ram 	: in  std_logic_vector(0 to address_width_ram-1)
        );
    end component;
    

   --Inputs
   signal clock_internal 				: std_logic := '0';
   signal data_in_internal 			: std_logic_vector(0 to data_width_testbench-1) := (others => '0');
   signal write_enable_internal 		: std_logic := '0';
   signal write_address_internal 	: std_logic_vector(0 to address_width_testbench-1) := (others => '0');
   signal read_enable_internal 		: std_logic := '0';
   signal read_address_internal 		: std_logic_vector(0 to address_width_testbench-1) := (others => '0');

 	--Outputs
   signal data_out_internal	 		: std_logic_vector(0 to data_width_testbench-1);
	
   -- Clock period definition
   constant clock_period_testbench	: time := 10ns;
	-- State definitions
   type states is (idle, writing, reading, stop);
	signal state : states := idle;
	-- Address and data definitions
	signal address 						: natural range 0 to 2 ** address_width_testbench-1;
	signal data_in, data_out    		: natural range 0 to 2 ** data_width_testbench-1;
begin
 
	uut : ram port map
	(
          clock_ram 				=> clock_internal,
          data_in_ram 			=> data_in_internal,
          data_out_ram 			=> data_out_internal,
          write_enable_ram 	=> write_enable_internal,
          write_address_ram 	=> write_address_internal,
          read_enable_ram 		=> read_enable_internal,
          read_address_ram 	=> read_address_internal
   );

   clock_process :process
   begin
		clock_internal <= '0';
		wait for clock_period_testbench/2;
		clock_internal <= '1';
		wait for clock_period_testbench/2;
   end process;
 	
	stimulus:process (clock_internal)
	begin
		if rising_edge(clock_internal)
		then
			case state is
				when idle => 		write_enable_internal 	<= '1';
										read_enable_internal 	<= '1';
										address						<= 0;
										data_in						<= 0;
										state <= writing;
				when writing =>	if address < 2 ** address_width_testbench
										then							
											-- Convert natural to std_logic_vector
											write_address_internal 	<= std_logic_vector(to_unsigned(address,address_width_testbench));
											-- Convert natural to std_logic_vector
											data_in_internal 	<= std_logic_vector(to_unsigned(data_in,data_width_testbench));
											-- Change state
											state <= reading;
										else
											-- Stop the machine
											state <= stop;
										end if;
				when reading =>	if address < 2 ** address_width_testbench
										then
											-- Convert natural to std_logic_vector
											read_address_internal  <= std_logic_vector(to_unsigned(address,address_width_testbench));
											-- Convert std_logic_vector to natural
											data_out <= to_integer(unsigned(data_out_internal));
											-- Check data
											assert data_out = data_in report "Non consistent data";
											-- Increment address
											address <= address+1;
											-- Increment data
											-- Suppress warning "vector truncated" by using a If for modulo operation
											if data_in = 2**data_width_testbench-1
											then 
												data_in <= 0;
											else
												data_in <= data_in+1;
											end if;
											-- Change state
											state <= writing;
										else
											-- Stop the machine
											state <= stop;
										end if;
				when stop =>		null;
			end case;
		end if;
	
	end process;
  
END;
